import typing
from enum import Enum
from urllib.parse import unquote_plus

import multipart
from multipart.multipart import parse_options_header

from renus.core.datastructures import FormData, UploadFile


class FormMessage(Enum):
    FIELD_START = 1
    FIELD_NAME = 2
    FIELD_DATA = 3
    FIELD_END = 4
    END = 5


class MultiPartMessage(Enum):
    PART_BEGIN = 1
    PART_DATA = 2
    PART_END = 3
    HEADER_FIELD = 4
    HEADER_VALUE = 5
    HEADER_END = 6
    HEADERS_FINISHED = 7
    END = 8


def _user_safe_decode(src: bytes, codec: str) -> str:
    try:
        return src.decode(codec)
    except (UnicodeDecodeError, LookupError):
        return src.decode("utf-8")


class FormParser:
    def __init__(
        self, headers, stream: typing.AsyncGenerator[bytes, None]
    ) -> None:
        self.headers = headers
        self.stream = stream
        self.messages = []  # type: typing.List[typing.Tuple[FormMessage, bytes]]

    def on_field_start(self) -> None:
        message = (FormMessage.FIELD_START, b"")
        self.messages.append(message)

    def on_field_name(self, data: bytes, start: int, end: int) -> None:
        message = (FormMessage.FIELD_NAME, data[start:end])
        self.messages.append(message)

    def on_field_data(self, data: bytes, start: int, end: int) -> None:
        message = (FormMessage.FIELD_DATA, data[start:end])
        self.messages.append(message)

    def on_field_end(self) -> None:
        message = (FormMessage.FIELD_END, b"")
        self.messages.append(message)

    def on_end(self) -> None:
        message = (FormMessage.END, b"")
        self.messages.append(message)

    async def parse(self) -> FormData:
        # Callbacks dictionary.
        callbacks = {
            "on_field_start": self.on_field_start,
            "on_field_name": self.on_field_name,
            "on_field_data": self.on_field_data,
            "on_field_end": self.on_field_end,
            "on_end": self.on_end,
        }

        # Create the parser.
        parser = multipart.QuerystringParser(callbacks)
        field_name = b""
        field_value = b""

        items = (
            []
        )  # type: typing.List[typing.Tuple[str, typing.Union[str, UploadFile]]]

        # Feed the parser with data from the request.
        async for chunk in self.stream:
            if chunk:
                parser.write(chunk)
            else:
                parser.finalize()
            messages = list(self.messages)
            self.messages.clear()
            for message_type, message_bytes in messages:
                if message_type == FormMessage.FIELD_START:
                    field_name = b""
                    field_value = b""
                elif message_type == FormMessage.FIELD_NAME:
                    field_name += message_bytes
                elif message_type == FormMessage.FIELD_DATA:
                    field_value += message_bytes
                elif message_type == FormMessage.FIELD_END:
                    name = unquote_plus(field_name.decode("utf-8"))
                    value = unquote_plus(field_value.decode("utf-8"))
                    items.append((name, value))
                elif message_type == FormMessage.END:
                    pass

        return FormData(items)


class MultiPartParser:
    def __init__(
        self, headers, stream: typing.AsyncGenerator[bytes, None]
    ) -> None:
        self.headers = headers
        self.stream = stream
        self.messages = []  # type: typing.List[typing.Tuple[MultiPartMessage, bytes]]

    def on_part_begin(self) -> None:
        message = (MultiPartMessage.PART_BEGIN, b"")
        self.messages.append(message)

    def on_part_data(self, data: bytes, start: int, end: int) -> None:
        message = (MultiPartMessage.PART_DATA, data[start:end])
        self.messages.append(message)

    def on_part_end(self) -> None:
        message = (MultiPartMessage.PART_END, b"")
        self.messages.append(message)

    def on_header_field(self, data: bytes, start: int, end: int) -> None:
        message = (MultiPartMessage.HEADER_FIELD, data[start:end])
        self.messages.append(message)

    def on_header_value(self, data: bytes, start: int, end: int) -> None:
        message = (MultiPartMessage.HEADER_VALUE, data[start:end])
        self.messages.append(message)

    def on_header_end(self) -> None:
        message = (MultiPartMessage.HEADER_END, b"")
        self.messages.append(message)

    def on_headers_finished(self) -> None:
        message = (MultiPartMessage.HEADERS_FINISHED, b"")
        self.messages.append(message)

    def on_end(self) -> None:
        message = (MultiPartMessage.END, b"")
        self.messages.append(message)

    async def parse(self) -> FormData:
        # Parse the Content-Type header to get the multipart boundary.
        content_type, params = parse_options_header(self.headers["content-type"])
        charset = params.get(b"charset", "utf-8")
        if type(charset) == bytes:
            charset = charset.decode("utf-8")
        boundary = params.get(b"boundary")

        # Callbacks dictionary.
        callbacks = {
            "on_part_begin": self.on_part_begin,
            "on_part_data": self.on_part_data,
            "on_part_end": self.on_part_end,
            "on_header_field": self.on_header_field,
            "on_header_value": self.on_header_value,
            "on_header_end": self.on_header_end,
            "on_headers_finished": self.on_headers_finished,
            "on_end": self.on_end,
        }

        # Create the parser.
        parser = multipart.MultipartParser(boundary, callbacks)
        header_field = b""
        header_value = b""
        content_disposition = None
        content_type = b""
        field_name = ""
        data = b""
        file = None  # type: typing.Optional[UploadFile]

        items = (
            []
        )  # type: typing.List[typing.Tuple[str, typing.Union[str, UploadFile]]]

        # Feed the parser with data from the request.
        async for chunk in self.stream:
            parser.write(chunk)
            messages = list(self.messages)
            self.messages.clear()
            for message_type, message_bytes in messages:
                if message_type == MultiPartMessage.PART_BEGIN:
                    content_disposition = None
                    content_type = b""
                    data = b""
                elif message_type == MultiPartMessage.HEADER_FIELD:
                    header_field += message_bytes
                elif message_type == MultiPartMessage.HEADER_VALUE:
                    header_value += message_bytes
                elif message_type == MultiPartMessage.HEADER_END:
                    field = header_field.lower()
                    if field == b"content-disposition":
                        content_disposition = header_value
                    elif field == b"content-type":
                        content_type = header_value
                    header_field = b""
                    header_value = b""
                elif message_type == MultiPartMessage.HEADERS_FINISHED:
                    disposition, options = parse_options_header(content_disposition)
                    field_name = _user_safe_decode(options[b"name"], charset)
                    if b"filename" in options:
                        filename = _user_safe_decode(options[b"filename"], charset)
                        file = UploadFile(
                            filename=filename,
                            content_type=content_type.decode("utf-8"),
                        )
                    else:
                        file = None
                elif message_type == MultiPartMessage.PART_DATA:
                    if file is None:
                        data += message_bytes
                    else:
                        await file.write(message_bytes)
                elif message_type == MultiPartMessage.PART_END:
                    if file is None:
                        items.append((field_name, _user_safe_decode(data, charset)))
                    else:
                        await file.seek(0)
                        items.append((field_name, file))
                elif message_type == MultiPartMessage.END:
                    pass

        parser.finalize()
        return FormData(items)



import functools
import typing
from typing import Any, AsyncGenerator, Iterator

import anyio

T = typing.TypeVar("T")


async def run_in_threadpool(
        func: typing.Callable[..., T], *args: typing.Any, **kwargs: typing.Any
) -> T:
    if kwargs:  # pragma: no cover
        # run_sync doesn't accept 'kwargs', so bind them in here
        func = functools.partial(func, **kwargs)
    return await anyio.to_thread.run_sync(func, *args)


class _StopIteration(Exception):
    pass


def _next(iterator: Iterator) -> Any:
    # We can't raise `StopIteration` from within the threadpool iterator
    # and catch it outside that context, so we coerce them into a different
    # exception type.
    try:
        return next(iterator)
    except StopIteration:
        raise _StopIteration


async def iterate_in_threadpool(iterator: Iterator) -> AsyncGenerator:
    while True:
        try:
            yield await anyio.to_thread.run_sync(_next, iterator)
        except _StopIteration:
            break



import gzip
import http.cookies
import inspect
import io
import json
import os
import stat
import typing
from email.utils import formatdate
from functools import partial
from hashlib import md5
from mimetypes import guess_type
from urllib.parse import quote, quote_plus

import anyio

from renus.core.cache import Cache
from renus.core.concurrency import iterate_in_threadpool
from renus.core.crypt import shift
from renus.core.datastructures import Background
from renus.core.serialize import jsonEncoder
from renus.core.status import Status
from renus.util.helper import get_random_string


class Response:
    media_type = "text/html"
    charset = "utf-8"

    def __init__(
            self,
            content: typing.Any = None,
            status_code: Status = Status.HTTP_200_OK,
            headers: dict = None,
            media_type: str = None,
            background: Background = None,
    ) -> None:
        self.status_code = status_code
        if media_type is not None:
            self.media_type = media_type
        self.body = self.render(content)
        self.init_headers(headers)
        self.background = background

    def render(self, content: typing.Any) -> bytes:
        if content is None:
            return b""
        if isinstance(content, bytes):
            return content
        return content.encode(self.charset)

    def init_headers(self, headers: typing.Mapping[str, str] = None) -> None:
        populate_content_type = True
        raw_headers = []
        if headers is not None:
            for k, v in headers.items():
                if k.lower() != 'content-length':
                    raw_headers.append((k.lower().encode("utf-8"), v.encode("utf-8")))

                if k.lower() == 'content-type':
                    populate_content_type = False

        content_type = self.media_type
        if content_type is not None and populate_content_type:
            if content_type.startswith("text/"):
                content_type += "; charset=" + self.charset
            raw_headers.append((b"content-type", content_type.encode("utf-8")))

        self.raw_headers = raw_headers

    def set_cookie(
            self,
            key: str,
            value: str = "",
            max_age: int = None,
            expires: int = None,
            path: str = "/",
            domain: str = None,
            secure: bool = False,
            httponly: bool = False,
            samesite: str = "lax",
    ) -> None:
        cookie = http.cookies.SimpleCookie()  # type: http.cookies.BaseCookie
        cookie[key] = value
        if max_age is not None:
            cookie[key]["max-age"] = max_age
        if expires is not None:
            cookie[key]["expires"] = expires
        if path is not None:
            cookie[key]["path"] = path
        if domain is not None:
            cookie[key]["domain"] = domain
        if secure:
            cookie[key]["secure"] = True
        if httponly:
            cookie[key]["httponly"] = True
        if samesite is not None:
            assert samesite.lower() in [
                "strict",
                "lax",
                "none",
            ], "samesite must be either 'strict', 'lax' or 'none'"
            cookie[key]["samesite"] = samesite
        cookie_val = cookie.output(header="").strip()
        self.raw_headers.append((b"set-cookie", cookie_val.encode("utf-8")))

    def delete_cookie(self, key: str, path: str = "/", domain: str = None) -> None:
        self.set_cookie(key, expires=0, max_age=0, path=path, domain=domain)

    async def __call__(self, request, scope, receive, send) -> None:
        headers = request.headers
        if "gzip" in headers.get("accept-encoding", "") and len(self.body) > 500:
            self.gzip_buffer = io.BytesIO()
            self.gzip_file = gzip.GzipFile(mode="wb", fileobj=self.gzip_buffer)
            self.gzip_file.write(self.body)
            self.gzip_file.close()
            self.body = self.gzip_buffer.getvalue()
            self.raw_headers.append((b"content-encoding", "gzip".encode("utf-8")))
        else:
            self.raw_headers.append((b"content-encoding", "none".encode("utf-8")))
        self.raw_headers.append((b"content-length", str(len(self.body)).encode("utf-8")))
        await send(
            {
                "type": "http.response.start",
                "status": self.status_code,
                "headers": self.raw_headers,
            }
        )
        await send({"type": "http.response.body", "body": self.body})

        if self.background is not None:
            await self.background()


class HtmlResponse(Response):
    media_type = "text/html"


class TextResponse(Response):
    media_type = "text/plain"

    def render(self, content: typing.Any) -> bytes:
        if content is None:
            return b""
        if isinstance(content, bytes):
            return content
        return str(content).encode(self.charset)


class JsonResponse(Response):
    media_type = "application/json"

    def render(self, content: typing.Any) -> bytes:
        if isinstance(content, bytes):
            return content

        return json.dumps(
            content,
            ensure_ascii=False,
            allow_nan=True,
            indent=None,
            separators=(",", ":"),
            cls=jsonEncoder
        ).encode("utf-8")


class JsonResponseRedirect(Response):

    def __init__(self, content: typing.Any = None, headers: dict = None,
                 background: Background = None) -> None:
        super().__init__(content, 307, headers, "application/json", background)

    def render(self, url: str) -> bytes:
        content = {"location": quote_plus(url, safe=":/%#?&=@[]!$&'()*+,;")}

        return json.dumps(
            content,
            ensure_ascii=False,
            allow_nan=False,
            indent=None,
            separators=(",", ":"),
            cls=jsonEncoder
        ).encode("utf-8")


class EncryptResponse(Response):
    media_type = "text/plain"

    def __init__(self, content: typing.Any, password: str, status_code: Status = Status.HTTP_200_OK,
                 headers: dict = None,
                 media_type: str = None, background: Background = None) -> None:
        self.password = password
        if headers is None:
            headers = {}

        headers["encrypted"] = '1'
        headers["encrypted_type"] = str(type(content))
        super().__init__(content, status_code, headers, media_type, background)

    def render(self, content: typing.Any) -> bytes:
        if isinstance(content, bytes):
            return content

        return shift(str(content), self.password).encode("utf-8")


class RedirectResponse(Response):
    def __init__(
            self,
            url: str,
            status_code: Status = Status.HTTP_307_TEMPORARY_REDIRECT,
            headers: dict = None,
            background: Background = None,
    ) -> None:
        if headers is None:
            headers = {}

        headers["location"] = quote_plus(url, safe=":/%#?&=@[]!$&'()*+,;")
        super().__init__(
            content=b"", status_code=status_code, headers=headers, background=background
        )


class StreamingResponse(Response):
    def __init__(
            self,
            content: typing.Any,
            status_code: Status = Status.HTTP_200_OK,
            headers: dict = None,
            media_type: str = None,
            background: Background = None,
            zipped: bool = True
    ) -> None:
        super().__init__()
        if inspect.isasyncgen(content):
            self.body_iterator = content
        else:
            self.body_iterator = iterate_in_threadpool(content)
        self.status_code = status_code
        self.media_type = self.media_type if media_type is None else media_type
        self.background = background
        self.zipped = zipped
        self.init_headers(headers)

    async def listen_for_disconnect(self, receive) -> None:
        while True:
            message = await receive()
            if message["type"] == "http.disconnect":
                break

    async def stream_response(self, request, scope, send) -> None:
        headers = request.headers
        is_zip = False
        if self.zipped and "gzip" in headers.get("accept-encoding", ""):
            is_zip = True
            self.gzip_buffer = io.BytesIO()
            self.gzip_file = gzip.GzipFile(mode="wb", fileobj=self.gzip_buffer)
            self.raw_headers.append((b"content-encoding", "gzip".encode("utf-8")))
        else:
            self.raw_headers.append((b"content-encoding", "none".encode("utf-8")))

        await send(
            {
                "type": "http.response.start",
                "status": self.status_code,
                "headers": self.raw_headers,
            }
        )
        async for chunk in self.body_iterator:
            if is_zip:
                self.gzip_file.write(chunk)
                chunk = self.gzip_buffer.getvalue()
                self.gzip_buffer.seek(0)
                self.gzip_buffer.truncate()

            if not isinstance(chunk, bytes):
                chunk = chunk.encode(self.charset)

            await send({"type": "http.response.body", "body": chunk, "more_body": True})
        chunk = b""
        if is_zip:
            self.gzip_file.write(chunk)
            self.gzip_file.close()
            chunk = self.gzip_buffer.getvalue()
            self.gzip_buffer.seek(0)
            self.gzip_buffer.truncate()
        await send({"type": "http.response.body", "body": chunk, "more_body": False})

    async def __call__(self, request, scope, receive, send) -> None:
        async with anyio.create_task_group() as task_group:
            async def wrap(func: "typing.Callable[[], typing.Awaitable[None]]") -> None:
                await func()
                task_group.cancel_scope.cancel()

            task_group.start_soon(wrap, partial(self.stream_response, request, scope, send))
            await wrap(partial(self.listen_for_disconnect, receive))
        if self.background is not None:
            await self.background()


class FileResponse(Response):
    chunk_size = 4096

    def __init__(
            self,
            path: str,
            status_code: Status = Status.HTTP_200_OK,
            headers: dict = None,
            media_type: str = None,
            background: Background = None,
            filename: str = None,
            stat_result: os.stat_result = None,
            method: str = None,
            zipped: bool = True,
            on_end=None
    ) -> None:
        super().__init__()
        self.path = path
        self.status_code = status_code
        self.background = background
        self.filename = filename
        self.on_end = on_end
        self.zipped = zipped
        self.send_header_only = method is not None and method.upper() == "HEAD"
        if media_type is None:
            media_type = guess_type(filename or path)[0] or "text/plain"
        self.media_type = media_type
        self.init_headers(headers)
        if self.filename is not None:
            content_disposition_filename = quote(self.filename)
            if content_disposition_filename != self.filename:
                content_disposition = "attachment; filename*=utf-8''{}".format(
                    content_disposition_filename
                )
            else:
                content_disposition = 'attachment; filename="{}"'.format(self.filename)
            self.raw_headers.append((b"content-disposition", content_disposition.lower().encode("utf-8")))

        self.stat_result = stat_result
        if stat_result is not None:
            self.set_stat_headers(stat_result, False)

    def set_stat_headers(self, stat_result: os.stat_result, is_zip) -> None:
        content_length = str(stat_result.st_size)
        last_modified = formatdate(stat_result.st_mtime, usegmt=True)
        etag_base = str(stat_result.st_mtime) + "-" + str(stat_result.st_size)
        etag = md5(etag_base.encode()).hexdigest()
        if not is_zip:
            self.raw_headers.append((b"content-length", content_length.lower().encode("utf-8")))
        self.raw_headers.append((b"last-modified", last_modified.lower().encode("utf-8")))
        self.raw_headers.append((b"etag", etag.lower().encode("utf-8")))

    async def __call__(self, request, scope, receive, send) -> None:
        headers = request.headers
        is_zip = False
        if self.zipped and "gzip" in headers.get("accept-encoding", "") and not self.send_header_only:
            is_zip = True
            self.gzip_buffer = io.BytesIO()
            self.gzip_file = gzip.GzipFile(mode="wb", fileobj=self.gzip_buffer)
            self.raw_headers.append((b"content-encoding", "gzip".encode("utf-8")))
        else:
            self.raw_headers.append((b"content-encoding", "none".encode("utf-8")))

        if self.stat_result is None:
            try:
                stat_result = await anyio.to_thread.run_sync(os.stat, self.path)
                self.set_stat_headers(stat_result, is_zip)
            except FileNotFoundError:
                raise RuntimeError(f"File at path {self.path} does not exist.")
            else:
                mode = stat_result.st_mode
                if not stat.S_ISREG(mode):
                    raise RuntimeError(f"File at path {self.path} is not a file.")

        await send(
            {
                "type": "http.response.start",
                "status": self.status_code,
                "headers": self.raw_headers,
            }
        )
        if self.send_header_only:
            await send({"type": "http.response.body", "body": b"", "more_body": False})
        else:
            async with await anyio.open_file(self.path, mode="rb") as file:
                more_body = True
                while more_body:
                    chunk = await file.read(self.chunk_size)
                    more_body = len(chunk) == self.chunk_size
                    if is_zip:
                        self.gzip_file.write(chunk)
                        if not more_body:
                            self.gzip_file.close()
                        chunk = self.gzip_buffer.getvalue()
                        self.gzip_buffer.seek(0)
                        self.gzip_buffer.truncate()

                    await send(
                        {
                            "type": "http.response.body",
                            "body": chunk,
                            "more_body": more_body,
                        }
                    )
        if self.on_end is not None:
            self.on_end()
        if self.background is not None:
            await self.background()


class PrivateResponse(Response):
    media_type = "text/plain"

    def __init__(
            self,
            url: str,
            status_code: Status = Status.HTTP_200_OK,
            headers: dict = None,
            background: Background = None
    ) -> None:
        super().__init__(
            content=url, status_code=status_code, headers=headers, background=background
        )

    def render(self, content: typing.Any) -> bytes:
        if content is None:
            return b""
        rnd = get_random_string(50)
        Cache().put(rnd, 'access to private')
        content = f'/storage/private/{rnd}/{content}'
        return content.encode(self.charset)



import inspect
import traceback

from renus.core.cache import Cache
from renus.core.concurrency import run_in_threadpool
from renus.core.config import Config
from renus.core.exception import debug_response
from renus.core.middleware import Middleware
from renus.core.request import Request
from renus.core.response import Response, TextResponse, JsonResponse
from renus.core.routing import Router
from renus.core.status import Status
from renus.core.websockets import WebSocket


class App:
    store = {}

    def __init__(self, lifespan=None,
                 on_startup=None,
                 on_shutdown=None, middlewares: list = None) -> None:
        self.debug = Config('app').get('debug', False)
        self.env = Config('app').get('env', 'local')
        self.on_startup = [] if on_startup is None else list(on_startup)
        self.on_shutdown = [] if on_shutdown is None else list(on_shutdown)
        self.middlewares = [] if middlewares is None else middlewares

        async def default_lifespan(app):
            await self.startup()
            yield
            await self.shutdown()

        self.lifespan_context = default_lifespan if lifespan is None else lifespan

    async def __call__(self, scope, receive, send) -> None:
        assert scope["type"] in ("http", "websocket", "lifespan")

        if scope["type"] == "http":
            await self.http(scope, receive, send)

        if scope["type"] == "lifespan":
            await self.lifespan(scope, receive, send)

        if scope["type"] == "websocket":
            await self.websocket(scope, receive, send)

    async def lifespan(self, scope, receive, send) -> None:
        """
        Handle ASGI lifespan messages, which allows us to manage application
        startup and shutdown events.
        """
        first = True
        app = scope.get("app")
        await receive()
        try:
            if inspect.isasyncgenfunction(self.lifespan_context):
                async for item in self.lifespan_context(app):
                    assert first, "Lifespan context yielded multiple times."
                    first = False
                    await send({"type": "lifespan.startup.complete"})
                    await receive()
            else:
                for item in self.lifespan_context(app):  # type: ignore
                    assert first, "Lifespan context yielded multiple times."
                    first = False
                    await send({"type": "lifespan.startup.complete"})
                    await receive()
        except BaseException:
            if first:
                exc_text = traceback.format_exc()
                await send({"type": "lifespan.startup.failed", "message": exc_text})
            if self.env == 'local':
                raise
        else:
            await send({"type": "lifespan.shutdown.complete"})

    async def websocket(self, scope, receive, send) -> None:
        scope["method"] = 'WS'
        ws = WebSocket(scope, receive, send)

        try:
            res = self.load_routes(ws, scope)
            setattr(ws, 'route', res)
            if not res:
                await ws.close(1004)
            else:
                middlewares = self.middlewares + res['middlewares']
                passed = Middleware(ws, middlewares).next()

                if passed is not True:
                    await ws.close(1003)
                    return

                if res["controller"] is not None:
                    if 'request' in inspect.getfullargspec(res["controller"].__init__).args:
                        controller = res["controller"](ws)
                    else:
                        controller = res["controller"]()
                    method = getattr(controller, res['func'])
                else:
                    method = res['func']

                res['args']['ws'] = ws

                await method(**res['args'])

        except Exception as exc:
            debug_response(exc, self.debug)
            if self.env == 'local':
                raise

    async def http(self, scope, receive, send):
        scope["method"] = scope["method"].upper()
        request = Request(scope, receive)

        if scope["method"] in ['POST', 'PUT', 'DELETE']:
            await request.form_safe()

        try:
            await self.view(request, scope, receive, send)
        except Exception as exc:
            debug = debug_response(exc, self.debug)
            await self.result(request, JsonResponse(*debug), scope, receive, send)
            if self.env == 'local':
                raise

    async def view(self, request, scope, receive, send):
        res = self.load_routes(request, scope)
        setattr(request, 'route', res)

        if not res:
            await self.result(request, TextResponse('Path not Found', Status.HTTP_404_NOT_FOUND), scope, receive, send)
        else:
            middlewares = self.middlewares + res['middlewares']
            passed = Middleware(request, middlewares).next()

            if passed is not True and request.method != 'OPTIONS':
                await self.result(request, JsonResponse(passed, Status.HTTP_403_FORBIDDEN), scope, receive, send)
            else:
                if res["controller"] is not None:
                    if 'request' in inspect.getfullargspec(res["controller"].__init__).args:
                        controller = res["controller"](request)
                    else:
                        controller = res["controller"]()
                    method = getattr(controller, res['func'])
                else:
                    method = res['func']

                if 'request' in method.__code__.co_varnames:
                    res['args']['request'] = request

                r = await self.function(res, request, method)

                await self.result(request, r, scope, receive, send)

    async def function(self, res, request, method):
        async def get_async():
            if inspect.iscoroutinefunction(method):
                return await method(**res['args'])
            else:
                return await run_in_threadpool(method, **res['args'])

        if res['cache']:
            c = Cache(use_hash=True).get(request.full_path, None)

            if c is None:
                r = await get_async()
                Cache(use_hash=True).put(request.full_path, r, res['cache'])
            else:
                r = c
                r.raw_headers.append((b'r-cache', b'ok'))
        else:
            r = await get_async()
        return r

    async def startup(self) -> None:
        """
        Run any `.on_startup` event handlers.
        """
        print('application startup')
        for handler in self.on_startup:
            if inspect.isasyncgenfunction(handler):
                await handler()
            else:
                handler()

    async def shutdown(self) -> None:
        """
        Run any `.on_shutdown` event handlers.
        """
        print('application shutdown')
        for handler in self.on_shutdown:
            if inspect.isasyncgenfunction(handler):
                await handler()
            else:
                handler()

    async def result(self, request, response, scope, receive, send):
        if not isinstance(response, Response):
            response = TextResponse(response)
        await response(request, scope, receive, send)

    def load_routes(self, req, scope):
        return Router(scope).response(req)



import asyncio
import tempfile
import typing

from renus.core.concurrency import run_in_threadpool


class Background:
    def __init__(
        self, func: typing.Callable, *args: typing.Any, **kwargs: typing.Any
    ) -> None:
        self.func = func
        self.args = args
        self.kwargs = kwargs
        self.is_async = asyncio.iscoroutinefunction(func)

    async def __call__(self) -> None:
        if self.is_async:
            await self.func(*self.args, **self.kwargs)
        else:
            await run_in_threadpool(self.func, *self.args, **self.kwargs)

class MultiDict(typing.Mapping):
    def __init__(
        self,
        *args: typing.Union[
            "MultiDict",
            typing.Mapping,
            typing.List[typing.Tuple[typing.Any, typing.Any]],
        ],
        **kwargs: typing.Any,
    ) -> None:
        assert len(args) < 2, "Too many arguments."

        if args:
            value = args[0]
        else:
            value = []

        if kwargs:
            value = (
                MultiDict(value).multi_items()
                + MultiDict(kwargs).multi_items()
            )

        if not value:
            _items = []  # type: typing.List[typing.Tuple[typing.Any, typing.Any]]
        elif hasattr(value, "multi_items"):
            value = typing.cast(MultiDict, value)
            _items = list(value.multi_items())
        elif hasattr(value, "items"):
            value = typing.cast(typing.Mapping, value)
            _items = list(value.items())
        else:
            value = typing.cast(
                typing.List[typing.Tuple[typing.Any, typing.Any]], value
            )
            _items = list(value)

        self._dict = {k: v for k, v in _items}
        self._list = _items

    def keys(self) -> typing.KeysView:
        return self._dict.keys()

    def values(self) -> typing.ValuesView:
        return self._dict.values()

    def items(self) -> typing.ItemsView:
        return self._dict.items()

    def multi_items(self) -> typing.List[typing.Tuple[str, str]]:
        return list(self._list)

    def get(self, key: typing.Any, default: typing.Any = None) -> typing.Any:
        if key in self._dict:
            return self._dict[key]
        return default

    def __getitem__(self, key: typing.Any) -> str:
        return self._dict[key]

    def __contains__(self, key: typing.Any) -> bool:
        return key in self._dict

    def __iter__(self) -> typing.Iterator[typing.Any]:
        return iter(self.keys())

    def __len__(self) -> int:
        return len(self._dict)

    def __eq__(self, other: typing.Any) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return sorted(self._list) == sorted(other._list)

    def __repr__(self) -> str:
        class_name = self.__class__.__name__
        items = self.multi_items()
        return f"{class_name}({items!r})"


class UploadFile:

    spool_max_size = 1024 * 1024

    def __init__(
        self, filename: str, file: typing.IO = None, content_type: str = ""
    ) -> None:
        self.filename = filename
        self.content_type = content_type
        if file is None:
            file = tempfile.SpooledTemporaryFile(max_size=self.spool_max_size)
        self.file = file

    @property
    def _in_memory(self) -> bool:
        rolled_to_disk = getattr(self.file, "_rolled", True)
        return not rolled_to_disk

    async def write(self, data: typing.Union[bytes, str]) -> None:
        if self._in_memory:
            self.file.write(data)  # type: ignore
        else:
            await run_in_threadpool(self.file.write, data)

    async def read(self, size: int = -1) -> typing.Union[bytes, str]:
        if self._in_memory:
            return self.file.read(size)
        return await run_in_threadpool(self.file.read, size)

    async def seek(self, offset: int) -> None:
        if self._in_memory:
            self.file.seek(offset)
        else:
            await run_in_threadpool(self.file.seek, offset)

    async def close(self) -> None:
        if self._in_memory:
            self.file.close()
        else:
            await run_in_threadpool(self.file.close)


class FormData(MultiDict):
    """
    An immutable multidict, containing both file uploads and text input.
    """

    def __init__(
            self,
            *args: typing.Union[
                "FormData",
                typing.Mapping[str, typing.Union[str, UploadFile]],
                typing.List[typing.Tuple[str, typing.Union[str, UploadFile]]],
            ],
            **kwargs: typing.Union[str, UploadFile],
    ) -> None:
        super().__init__(*args, **kwargs)

    async def close(self) -> None:
        for key, value in self.multi_items():
            if isinstance(value, UploadFile):
                await value.close()



from renus.core.config import Config

app = Config('app')


class Cprint:

    def __init__(self, debug: bool = None) -> None:
        """
        @param debug: by default read debug mode from Config file
        """
        self._debug = app.get('debug', False) if debug is None else debug

    def print(self, txt):
        """
        print txt if debug mode is True
        """
        if self._debug:
            print(txt)

    @staticmethod
    def bold(txt):
        return f'\033[1m{txt}\033[0m'

    @staticmethod
    def italic(txt):
        return f'\033[3m{txt}\033[0m'

    @staticmethod
    def underline(txt):
        return f'\033[4m{txt}\033[0m'

    @staticmethod
    def mid_line(txt):
        return f'\033[9m{txt}\033[0m'

    @staticmethod
    def black(txt):
        return f'\033[30m{txt}\033[0m'

    @staticmethod
    def grey(txt):
        return f'\033[37m{txt}\033[0m'

    @staticmethod
    def red(txt):
        return f'\033[91m{txt}\033[0m'

    @staticmethod
    def green(txt):
        return f'\033[92m{txt}\033[0m'

    @staticmethod
    def yellow(txt):
        return f'\033[93m{txt}\033[0m'

    @staticmethod
    def blue(txt):
        return f'\033[94m{txt}\033[0m'

    @staticmethod
    def pink(txt):
        return f'\033[95m{txt}\033[0m'

    @staticmethod
    def cyan(txt):
        return f'\033[96m{txt}\033[0m'

    @staticmethod
    def white(txt):
        return f'\033[97m{txt}\033[0m'

    @staticmethod
    def bg_white(txt):
        return f'\033[7m{txt}\033[0m'

    @staticmethod
    def bg_black(txt):
        return f'\033[40m{txt}\033[0m'

    @staticmethod
    def bg_red(txt):
        return f'\033[41m{txt}\033[0m'

    @staticmethod
    def bg_green(txt):
        return f'\033[42m{txt}\033[0m'

    @staticmethod
    def bg_yellow(txt):
        return f'\033[43m{txt}\033[0m'

    @staticmethod
    def bg_blue(txt):
        return f'\033[44m{txt}\033[0m'

    @staticmethod
    def bg_pink(txt):
        return f'\033[45m{txt}\033[0m'

    @staticmethod
    def bg_cyan(txt):
        return f'\033[46m{txt}\033[0m'

    @staticmethod
    def bg_grey(txt):
        return f'\033[47m{txt}\033[0m'



import os
from datetime import datetime

class Log:
    def __init__(self) -> None:
        self.filename='storage/logs/Renus-'+datetime.utcnow().strftime('%Y-%m-%d')+'.log'

    def info(self,msg):
        create(self.filename, msg, 'info')

    def warning(self,msg):
        create(self.filename, msg, 'warning')

    def error(self,msg):
        create(self.filename, msg, 'error')



def create(filename, value,typ):
    os.makedirs(os.path.dirname(filename), exist_ok=True)
    with open(filename, 'a') as output:
        output.write(f"[{datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S')}] {typ.upper()}: {str(value)}\n")


import typing
from datetime import datetime

from bson import ObjectId
from pymongo import MongoClient

from renus.core.config import Config
from renus.core.cprint import Cprint


class ReModel:
    def __init__(self, doc):
        for k, v in doc.items():
            setattr(self, k, v)

    def __iter__(self):
        return iter(self.__dict__.items())

    def __getitem__(self, item):
        return self.__dict__[item]


class ModelBase:
    client = MongoClient(Config('database').get('host', '127.0.0.1'),
                         Config('database').get('port', 27017),
                         username=Config('database').get('username', None),
                         password=Config('database').get('password', None))

    _database_name = Config('database').get('name', 'renus')
    collection_name = None
    metro = None
    storage = None
    hidden_fields = []
    document_model = dict
    add_time_fields = True

    def __init__(self) -> None:
        if hasattr(self, '_collection_name'):
            raise RuntimeError('_collection_name has expired. Use collection_name')

        self._steps = None
        self._where = None
        self._distinct = None
        self._limit = None
        self._skip = None
        self._sort = None
        self._select = None
        self.visible_fields = []

    def __iter__(self):
        return iter(self.get())

    def set_database(self, name: str):
        self._database_name = name
        return self

    def set_collection(self, name: str):
        self.collection_name = name
        return self

    @property
    def db(self):
        return self.client[self._database_name]

    def collection(self, name: str = None):
        if name is None:
            name = self.collection_name

        return self.db[name]

    def reset(self):
        self._steps = None
        self._where = None
        self._distinct = None
        self._limit = None
        self._skip = None
        self._sort = None
        self._select = None
        self.visible_fields = []
        return self

    def aggregate(self, pipeline: typing.Any, session: typing.Any = None):
        return self.__police(self.collection().aggregate(pipeline, session))

    def where(self, where: dict):
        if self._steps is not None:
            self._steps.append({
                "$match": where
            })
        else:
            if self._where is None:
                self._where = {}
            for k, v in where.items():
                if k.startswith('$') and k in self._where:
                    if type(v) is list:
                        self._where[k] += v
                    else:
                        self._where[k] = {**self._where[k], **v}
                else:
                    self._where[k] = v

        return self

    def count(self, skip=False, limit=False):
        k = {}
        if self._where is not None:
            k['filter'] = self._where
        else:
            k['filter'] = {}
        if skip and self._skip is not None:
            k['skip'] = self._skip
        if limit and self._limit is not None:
            k['limit'] = self._limit

        return self.collection().count_documents(**k)

    def distinct(self, key: str):
        self._distinct = key
        return self

    def limit(self, count: int):
        if self._steps is not None:
            self._steps.append({"$limit": count})
        else:
            self._limit = count
        return self

    def skip(self, count: int):
        if self._steps is not None:
            self._steps.append({"$skip": count})
        else:
            self._skip = count
        return self

    def group(self, item: dict):
        if self._steps is None:
            raise RuntimeError("group work with steps() architect")
        self._steps.append({"$group": item})
        return self

    def sort(self, key_or_list: typing.Union[str, typing.List[typing.Tuple]], asc: bool = True):
        if type(key_or_list) is dict:
            s = key_or_list
        elif type(key_or_list) is list:
            s = []
            for f, a in key_or_list:
                s.append((f, 1 if a else -1))
        else:
            s = [(key_or_list, 1 if asc else -1)]
        if self._steps is not None:
            if type(s) is list:
                s = {i[0]: i[1] for i in s}
            self._steps.append({"$sort": s})
        else:
            self._sort = s
        return self

    def select(self, *select: [str, typing.Dict]):
        s = {}
        if len(select) == 1 and type(select[0]) is dict:
            s = select[0]
        else:
            for key in select:
                if type(key) is dict:
                    s = key
                else:
                    s[key] = 1

        if self._steps is not None:
            self._steps.append({'$project': s})
            return self

        self._select = s
        return self

    def with_relation(self, collection, local_field, forigen_field, to, single=False, pipeline=None):
        if pipeline is None:
            pipeline = []
        if self._steps is not None:
            self._steps.append({
                "$lookup": {
                    "from": collection,
                    "localField": local_field,
                    "foreignField": forigen_field,
                    "as": to,
                    "pipeline": pipeline
                }
            })
            if single:
                self._steps.append({'$addFields': {f'{to}': {'$arrayElemAt': [f"${to}", 0]}}})
        else:
            c = Cprint()
            c.print(c.red('use steps mode for with_relation.'))
        return self

    def steps(self):
        self._steps = []
        return self

    def _get(self, police: bool = True, session=None) -> typing.List[document_model]:
        if self._steps is not None:
            return self.aggregate(self._steps, session)

        where, select = self._base_gate(police)

        find = self.collection().find(where, select, session=session)

        if self._sort is not None:
            find = find.sort(self._sort)
        if self._skip is not None:
            find = find.skip(self._skip)
        if self._limit is not None:
            find = find.limit(self._limit)
        if self._distinct is not None:
            return find.distinct(self._distinct)

        return self.__police(find) if police else list(find)

    def get(self, police: bool = True, session=None) -> typing.List[document_model]:
        return self._get(police, session)

    def _first(self, police: bool = True, session=None) -> typing.Union[document_model, None]:
        self.limit(1)
        res = self.get(police, session)
        self._limit = None
        if len(res) == 1:
            return res[0]
        return None

    def first(self, police: bool = True, session=None) -> typing.Union[document_model, None]:
        return self._first(police, session)

    def create(self, document: dict, session=None) -> dict:
        if self.add_time_fields and "updated_at" not in document:
            document["updated_at"] = datetime.utcnow()

        if self.add_time_fields and "created_at" not in document:
            document["created_at"] = datetime.utcnow()

        id = self.collection().insert_one(document, session=session).inserted_id
        document['_id'] = id
        self.boot_event('create', {}, document)
        self._attach_file(document)
        return document

    def create_many(self, documents: typing.List, session=None) -> typing.List[ObjectId]:
        if self.add_time_fields:
            for document in documents:
                if "updated_at" not in document:
                    document["updated_at"] = datetime.utcnow()

                if "created_at" not in document:
                    document["created_at"] = datetime.utcnow()

        ids = self.collection().insert_many(documents, session=session).inserted_ids
        self.boot_event('create_many', {}, ids)
        self._attach_file({'_id': 'create_many', 'documents': documents})
        return ids

    def update(self, new: dict, upsert=False, get_old=False, session=None) -> bool:
        where = self.__ud_gate('update')
        if self.add_time_fields:
            new["updated_at"] = datetime.utcnow()
        d = {"$set": new}
        if self.add_time_fields and "created_at" not in new:
            d["$setOnInsert"] = {'created_at': datetime.utcnow()}
        old = self.collection().find_one_and_update(where, d, upsert=upsert, session=session)
        self.boot_event('update', old, new)
        if old is None:
            new = {'_id': 'upsert', 'doc': new}
        else:
            new['_id'] = old['_id']
        self._attach_file(new)
        return old if get_old else True

    def update_opt(self, new: dict, upsert=False, get_old=False, session=None) -> bool:
        where = self.__ud_gate('update')
        if '$set' not in new:
            new['$set'] = {}
        if '$setOnInsert' not in new:
            new['$setOnInsert'] = {}
        if self.add_time_fields:
            new['$set']["updated_at"] = datetime.utcnow()
            new['$setOnInsert']['created_at'] = datetime.utcnow()
        old = self.collection().find_one_and_update(where, new, upsert=upsert, session=session)
        self.boot_event('update', old, {k[1:]: v for k, v in new.items()})
        if old is None:
            new = {'_id': 'upsert', 'doc': new}
        else:
            new['_id'] = old['_id']
        self._attach_file(new)
        return old if get_old else True

    def update_opt_many(self, new: dict, upsert=False, get_old=False, session=None) -> bool:
        where = self.__ud_gate('update')
        if '$set' not in new:
            new['$set'] = {}
        if '$setOnInsert' not in new:
            new['$setOnInsert'] = {}
        if self.add_time_fields:
            new['$set']["updated_at"] = datetime.utcnow()
            new['$setOnInsert']['created_at'] = datetime.utcnow()
        old = self.collection().update_many(where, new, upsert=upsert, session=session).raw_result
        self.boot_event('update_many', old, {k[1:]: v for k, v in new.items()})
        self._attach_file({'_id': 'update_opt_many', 'documents': new})
        return old if get_old else True

    def update_many(self, new: dict, session=None) -> bool:
        where = self.__ud_gate('update')
        if self.add_time_fields:
            new["updated_at"] = datetime.utcnow()
        old = self.collection().update_many(where, {"$set": new}, session=session).raw_result
        self.boot_event('update_many', old, str(new))
        self._attach_file({'_id': 'update_many', 'documents': new})
        return True

    def delete(self, all: bool = False, session=None) -> bool:
        where = self.__ud_gate('delete')
        if all is False:
            old = self.collection().find_one_and_delete(where, session=session)
            if self.metro is not None:
                self._handle_metro('delete', old)
            self.boot_event('delete', old, {})
        else:
            if self.metro is not None:
                self._handle_metro('delete')
            old = self.collection().delete_many(where, session=session)

            self.boot_event('delete_many', {'deleted_count': old.deleted_count, 'where': str(where)}, {})

        return old

    def make_visible(self, fields: typing.List):
        self.visible_fields = fields
        return self

    @staticmethod
    def boot_event(typ: str, old, new):
        pass

    @staticmethod
    def convert_id(id):
        if type(id) is str:
            try:
                return ObjectId(id)
            except Exception:
                return id
        return id

    def _base_gate(self, police) -> list:
        where = self._where
        select = self._select

        if select is None and police == False:
            raise RuntimeError("when police is off, for security reason select fields. ex: .select('_id','name')")

        if where is not None:
            if '_id' in where:
                where['_id'] = self.convert_id(where['_id'])

        return [where, select]

    @staticmethod
    def cast(document: dict):
        return document

    def __cleaner(self, document: dict):
        for field in self.hidden_fields:
            if field in document and field not in self.visible_fields:
                del document[field]

        return self.document_model(self.cast(document))

    def __police(self, documents: typing.Any):
        if documents is None:
            return None
        res = []
        if type(documents) is dict:
            return self.__cleaner(documents)
        else:
            for document in documents:
                if type(document) is dict:
                    res.append(self.__cleaner(document))
        return res

    def __ud_gate(self, type: str):
        where = self._where
        if where is None:
            raise RuntimeError(f"for {type} use where. ex: .where({'name':'test'})")
        if '_id' in where:
            where['_id'] = self.convert_id(where['_id'])

        return where

    def _handle_metro(self, typ, obj=None):
        if typ == 'delete':
            if obj is None:
                where = self.__ud_gate('delete')
                all = self.collection().find(where)
            else:
                all = [obj]

            for item in all:
                for field, db in self.metro.items():
                    if type(db) is not list:
                        raise RuntimeError(
                            "metro format not true. ex: '_id':[{'collection':'test','field': 'test_id'}]")
                    for d in db:
                        strg = d.get('storage', False)
                        c = d.get('collection', False)
                        if c is not False:
                            if strg:
                                files = self.collection(c).find({
                                    d['field']: item[field]
                                })

                                for file in files:
                                    self._remove_file(file.get(strg, []))
                            self.collection(c).delete_many({
                                d['field']: item[field]
                            })
                        elif strg is not False:
                            l = self.__get_links(strg, item)
                            if l:
                                self._remove_file(l)

    def __get_links(self, field: str, doc: dict):
        item = doc.copy()
        p = field.split('.')
        r = None
        for i in p:
            if i not in item:
                return False
            r = item[i]
            item = item[i]
        return r

    def _attach_file(self, item):
        if self.storage is None or self.metro is None:
            return
        links = self._links_extractor(item)
        self.storage.reset().where({
            'path': {'$in': links}
        }).update_many({'type': self.collection_name, 'type_id': item['_id']})

    def _remove_file(self, links):
        if self.storage is None:
            return
        if type(links) is dict:
            links = list(links.values())
        if type(links) is not list:
            links = [links]

        for link in links:
            if type(link) is str:
                self.storage.remove(link)
            elif type(link) is dict:
                self.storage.remove(link.get('url'))
            else:
                raise RuntimeError(f"type {link} must be string or dict but its {type(link)}")

    @property
    def database_name(self):
        return self._database_name

    def _links_extractor(self, item) -> list:
        s = []
        for field, db in self.metro.items():
            if type(db) is not list:
                raise RuntimeError(
                    "metro format not true. ex: '_id':[{'collection':'test','field': 'test_id'}]")
            for d in db:
                strg = d.get('storage', False)
                if strg is not False:
                    l = self.__get_links(strg, item)
                    if l:
                        s.extend(l)

        return s
